import java.util.Vector;

public class Tmports
   {

// Sends and receives data from ports on a circuit in the DE-3

TM tm;
Vector<String> portNames;
Vector<Integer> portDirs;
Vector<Integer> portWidths;

public final static int READ = 0;
public final static int WRITE = 1;

public Tmports() throws java.io.IOException {
   this(null);
}



public Tmports(String hostname) throws java.io.IOException {

   // Create a Tmports object.  Try connecting to the board on hostname.

   tm = new TM(hostname);
   portNames = new Vector<String>();
   portDirs = new Vector<Integer>();
   portWidths = new Vector<Integer>();
}



public void describePort(String portName, int direction, int portWidthBits) {

   // Tell Tmports about this port, so open() can recognize ports by
   // their names.
   // XXX -- you have to call describePort in the same order that they are
   // listed in the *.ports file, so that it get the port numbers right.
   // Maybe this routine should take the port number as an argument instead.
   // XXX -- direction should be an enum or something.  Braiden, I've forgotten
   // how to export enums in Java...

   if(portNames.contains(portName)) {
      System.out.println("Tmports.describePort() ignoring duplicate port named:" + portName);
      return;
   }

   if(direction != Tmports.READ && direction != Tmports.WRITE) {
      System.out.println("Tmports.describePort() illegal direction for port named:" + portName + " : " + direction);
      return;
   }

   portNames.add(portName);
   portDirs.add(new Integer(direction));
   portWidths.add(new Integer(portWidthBits));
}



public int open(String portName, String rw) {

   // Open this port, and return a descriptor that refers to it
   // Returns -1 on failure.

   int result = portNames.indexOf(portName);

   if(result == -1) {
      return(result);
   }

   int direction = portDirs.elementAt(result);

   if(rw.equals("r") && direction != Tmports.READ) {
      System.out.println("Tmports.open() illegal direction for port named:" + portName + " : " + direction);
      return(-1);
   }

   if(rw.equals("w") && direction != Tmports.WRITE) {
      System.out.println("Tmports.open() illegal direction for port named:" + portName + " : " + direction);
      return(-1);
   }

   return(result);
}



public boolean close(int port) {

   // Let people close ports

   if(port < 0 || port >= portNames.size()) {
      return(false);
   }

   return(true);
}



public int read(int port, 
          byte[] b,
          int off,
          int len) throws java.io.IOException {

   if(port < 0 || port >= portNames.size()) {
      return(-1);
   }

   if(portDirs.elementAt(port) != Tmports.READ) {
      return(-1);
   }

   int portWidthBits = portWidths.elementAt(port);
   int result = readPacket(b, off, len, port, portWidthBits);

   // Don't bother with byte swapping.  We'll use the
   // byte swapping in java.nio.* for now, and hope
   // that most people use standard Java types

	//    // The ports wrapper in the circuit returns its data in little endian order.
	//    // Java uses big endian (maybe only on X86 ??).
	//    // Swap the bytes in each value that was read, so that Java will
	//    // understand them.

	//    if(result <= 0) {
	//       return(result);
	//    }
	// 
	//    int portBytes = (portWidthBits + 7) / 8;
	//    int numValues = result / portBytes;
	//    byte[] temp = new byte[portBytes];

	// 
	//    for(int i = 0; i < numValues; i++) {
	//       for(int j = 0; j < portBytes; j++) {
	//          temp[j] = b[off + i * portBytes + j];
	//       }
	//       for(int j = 0; j < portBytes; j++) {
	//          b[off + i * portBytes + portBytes - 1 - j] = temp[j];
	//       }
	//    }

   return(result);
}



public int write(int port, 
          byte[] b,
          int off,
          int len) throws java.io.IOException {

   if(port < 0 || port >= portNames.size()) {
      return(-1);
   }

   if(portDirs.elementAt(port) != Tmports.WRITE) {
      return(-1);
   }

   int portWidthBits = portWidths.elementAt(port);

   // Don't bother with byte swapping.  We'll use the
   // byte swapping in java.nio.* for now, and hope
   // that most people use standard Java types

	//    int portBytes = (portWidthBits + 7) / 8;
	//    int numValues = len / portBytes;
	//    byte[] temp = new byte[portBytes];
	// 
	//    // The ports wrapper in the circuit wants its data in little endian order.
	//    // Java uses big endian (maybe only on X86 ??).
	//    // Swap the bytes in each value to be written, so that Java will
	//    // understand them.
	// 
	//    // XXX This totally trashes the user's data buffer.  We could:
	//    // a) live with it
	//    // b) alloc some space and copy the data before swapping the bytes
	//    // c) swap it back after we've written it
	//    // Braiden, what do you think ?
	// 
	//    for(int i = 0; i < numValues; i++) {
	//       for(int j = 0; j < portBytes; j++) {
	//          temp[j] = b[off + i * portBytes + j];
	//       }
	//       for(int j = 0; j < portBytes; j++) {
	//          b[off + i * portBytes + portBytes - 1 - j] = temp[j];
	//       }
	//    }

   writePacket(b, off, len, port, portWidthBits);

   return(len);
}

   



final static int TM_MAX_PACKET_DATA = 16383;
final static int TM_PACKET_WRITE =                     0x80000000;
final static int TM_PACKET_READ =                      0x00000000;
final static int TM_PACKET_NEEDS_ACK =                 0x40000000;
final static int TM_PACKET_PORT_MASK =                 0x03FC0000;
final static int TM_PACKET_LENGTH_LOW_BITS_MASK =      0x00000FFF;
final static int TM_PACKET_LENGTH_HIGH_BITS_MASK =     0x0C000000;
final static int TM_PACKET_PORT_WIDTH_HIGH_BITS_MASK = 0x0000F000;
final static int TM_PACKET_PORT_WIDTH_LOW_BITS_MASK =  0x00030000;


void writePacket(
          byte[] b,
          int off,
          int len,
          int port,
          int port_width_bits) throws java.io.IOException {

   // Send len bytes starting from b[off] to the given port

   int port_bytes = (port_width_bits + 7) / 8;

   while(len > 0) {

      int nbytes = len;

      if(nbytes > TM_MAX_PACKET_DATA) {
         nbytes = (TM_MAX_PACKET_DATA / port_bytes) * port_bytes;
      }
   
      // Write nbytes bytes to the given port
   
      int header = TM_PACKET_WRITE;
   
      // For now, the tmu ports package only supports hand-shaked ports.
      header |= TM_PACKET_NEEDS_ACK;
   
      header |= (port << 18) & TM_PACKET_PORT_MASK; 
   
      // The header bit fields are a mess, due to backwards compatibility
      header |= (nbytes & TM_PACKET_LENGTH_LOW_BITS_MASK);
      header |= ((nbytes << 14) & TM_PACKET_LENGTH_HIGH_BITS_MASK);
   
      int port_width_bytes = (port_width_bits + 7) / 8;
      
      header |= ((port_width_bytes << 10) & TM_PACKET_PORT_WIDTH_HIGH_BITS_MASK);
      header |= ((port_width_bytes << 16) & TM_PACKET_PORT_WIDTH_LOW_BITS_MASK);
   
      tm.writeData(intToByteArray(header), 0, 4);
      tm.writeData(b, off, nbytes);
   
      byte[] result_code_bytes = new byte[4];
   
      tm.readData(result_code_bytes, 0, 4);
   
      int result_code = byteArrayToInt(result_code_bytes);

      if(result_code == -1) {
         // XXX Need to throw an exception.
         return;
      }

      len -= result_code;
      off += result_code;
   }
}



int readPacket(
          byte[] b,
          int off,
          int len,
          int port,
          int port_width_bits) throws java.io.IOException {

   // Read len bytes from the given port into b[], starting at off

   int total_bytes = 0;

   while(len > 0) {
      // Read len bytes from the given port

      int nbytes = len;

      if(nbytes > TM_MAX_PACKET_DATA) {
         nbytes = TM_MAX_PACKET_DATA;
      }
   
      int header = TM_PACKET_READ;
   
      // For now, the tmu ports package only supports hand-shaked ports.
      header |= TM_PACKET_NEEDS_ACK;
   
      header |= (port << 18) & TM_PACKET_PORT_MASK; 
   
      // The header bit fields are a mess, due to backwards compatibility
      header |= (nbytes & TM_PACKET_LENGTH_LOW_BITS_MASK);
      header |= ((nbytes << 14) & TM_PACKET_LENGTH_HIGH_BITS_MASK);
   
      int port_width_bytes = (port_width_bits + 7) / 8;
      
      header |= ((port_width_bytes << 10) & TM_PACKET_PORT_WIDTH_HIGH_BITS_MASK);
      header |= ((port_width_bytes << 16) & TM_PACKET_PORT_WIDTH_LOW_BITS_MASK);
   
      tm.writeData(intToByteArray(header), 0, 4);
   
      byte[] result_code_bytes = new byte[4];
   
      int retval = tm.readData(result_code_bytes, 0, 4);
      if(retval != 4) {
         return (-1);
      }
   
      int result_code = byteArrayToInt(result_code_bytes);
      if(result_code != nbytes) {
         return (-1);
      }
   
      retval = tm.readData(b, off, nbytes);
      if(retval != nbytes ) {
         return (-1);
      }

      total_bytes += retval;
      len -= retval;
      off += retval;
   }

   return(total_bytes);
}



final byte[] intToByteArray(int value) {

   // Convert an integer to a byte array in big-endian order

   return new byte[] {
           (byte)(value >>> 24),
           (byte)(value >>> 16),
           (byte)(value >>> 8),
           (byte)value};
}


int byteArrayToInt(byte[] b) {

   // Create an int value from 4 bytes that are in big-endian order

   return(
	(((int)b[0] & 0xFF) << 24)
	| (((int)b[1] & 0xFF) << 16)
	| (((int)b[2] & 0xFF) << 8)
	| (((int)b[3] & 0xFF) << 0)
	);
}

}
